/*
 *
 *  Copyright (C) 2007  Uwe Weng
 *
 *  This file is part of Fuzzy Services, a library for processing fuzzy
 *  information.
 *
 *  Fuzzy Services are free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Fuzzy Services are distributed in the hope that they will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Fuzzy Services; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *  The license is also available at http://www.gnu.org/licenses/gpl.txt
 *
 ******************************************************************************/
package net.sourceforge.fuzzyservices.beans;

import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeSupport;

import java.beans.PropertyVetoException;
import java.io.Serializable;
import java.util.Iterator;
import java.util.logging.Level;
import java.util.logging.Logger;
import net.sourceforge.fuzzyservices.utils.FuzzyResourceManager;
import org.apache.commons.lang.SerializationUtils;
import org.apache.commons.lang.builder.EqualsBuilder;
import org.apache.commons.lang.builder.HashCodeBuilder;

/**
 * This class represents a fuzzy setDegreeOfMembership according to JavaBeans conventions.
 *
 * @see net.sourceforge.fuzzyservices.core.FuzzySet
 *
 * @version 1.0
 * @author Uwe Weng
 */
public class FuzzySet implements Serializable, PropertyChangeListener {

    /**
     * Default serial version UID.
     */
    private static final long serialVersionUID = 1L;
    /**
     * Unique technical identifier. Only for persistence is used.
     */
    private int id;
    //
    // Bound property names
    //
    /** Bound property name for <code>membershipFunction</code>. */
    public static final String MEMBERSHIP_FUNCTION_PROPERTY = "membershipFunction";
    /** Bound property name for <code>convex</code>. */
    public static final String CONVEX_PROPERTY = "convex";
    /** Bound property name for <code>height</code>. */
    public static final String HEIGHT_PROPERTY = "height";
    /** Bound property name for <code>normalized</code>. */
    public static final String NORMALIZED_PROPERTY = "normalized";
    /** Bound property name for <code>validFuzzyLRNumber</code>. */
    public static final String VALID_FUZZY_NUMBER_PROPERTY = "validFuzzyNumber";
    /** Bound property name for <code>validFuzzyLRNumber</code>. */
    public static final String VALID_FUZZY_LR_NUMBER_PROPERTY = "validFuzzyLRNumber";
    /** Bound property name for <code>validFuzzyInterval</code>. */
    public static final String VALID_FUZZY_INTERVAL_PROPERTY = "validFuzzyInterval";
    /** Bound property name for <code>validFuzzyLRInterval</code>. */
    public static final String VALID_FUZZY_LR_INTERVAL_PROPERTY = "validFuzzyLRInterval";
    /** The membership function property.*/
    private MembershipFunction membershipFunction = null;
    /** Support for any PropertyChangeListeners which have been registered. */
    private transient PropertyChangeSupport propertyChangeSupport = new PropertyChangeSupport(this);

    /** Default <code>FuzzySet</code> constructor. */
    public FuzzySet() {
        // Do nothing
    }

    /**
     * Creates a new <code>FuzzySet</code>.
     * @param newMembershipFunction membership function of the fuzzy set
     */
    public FuzzySet(final MembershipFunction newMembershipFunction) {
        this.membershipFunction = newMembershipFunction;
    }

    /**
     * Returns the technical identifier (e.g. within a database).
     * @return the ID
     */
    public int getId() {
        return id;
    }

    /**
     * Returns the membership function.
     * @return the <code>membershipFunction</code> property
     * @see #setMembershipFunction(MembershipFunction)
     */
    public final MembershipFunction getMembershipFunction() {
        return membershipFunction;
    }

    /**
     * Sets the membership function.
     * @param newMembershipFunction The new value for the property.
     * @see #getMembershipFunction()
     */
    public final void setMembershipFunction(final MembershipFunction newMembershipFunction) {
        MembershipFunction oldValue = this.membershipFunction;
        float oldHeight = getHeight();
        boolean oldIsConvex = isConvex();
        boolean oldIsNormalized = isNormalized();
        boolean oldIsValidFuzzyNumber = isValidFuzzyNumber();
        boolean oldIsValidFuzzyLRNumber = isValidFuzzyLRNumber();
        boolean oldIsValidFuzzyInterval = isValidFuzzyInterval();
        boolean oldIsValidFuzzyLRInterval = isValidFuzzyLRInterval();

        if (membershipFunction != null) {
            this.membershipFunction.removePropertyChangeListener(this);
        }
        this.membershipFunction = newMembershipFunction;
        if (membershipFunction != null) {
            this.membershipFunction.addPropertyChangeListener(this);
        }
        propertyChangeSupport.firePropertyChange(MEMBERSHIP_FUNCTION_PROPERTY,
                oldValue, newMembershipFunction);
        propertyChangeSupport.firePropertyChange(HEIGHT_PROPERTY,
                oldHeight, getHeight());
        propertyChangeSupport.firePropertyChange(CONVEX_PROPERTY, oldIsConvex,
                isConvex());
        propertyChangeSupport.firePropertyChange(NORMALIZED_PROPERTY,
                oldIsNormalized, isNormalized());
        propertyChangeSupport.firePropertyChange(VALID_FUZZY_NUMBER_PROPERTY,
                oldIsValidFuzzyNumber, isValidFuzzyNumber());
        propertyChangeSupport.firePropertyChange(VALID_FUZZY_LR_NUMBER_PROPERTY,
                oldIsValidFuzzyLRNumber, isValidFuzzyLRNumber());
        propertyChangeSupport.firePropertyChange(VALID_FUZZY_INTERVAL_PROPERTY,
                oldIsValidFuzzyInterval, isValidFuzzyInterval());
        propertyChangeSupport.firePropertyChange(VALID_FUZZY_LR_INTERVAL_PROPERTY,
                oldIsValidFuzzyLRInterval, isValidFuzzyLRInterval());
    }

    /**
     * Concentrates this fuzzy setDegreeOfMembership.
     * @exception PropertyVetoException when the attempt to set the property is
     * vetoed by a listener
     */
    public final void concentrate() throws PropertyVetoException {
        if (membershipFunction != null) {
            membershipFunction.concentrate();
        }
    }

    /**
     * Dilates this fuzzy setDegreeOfMembership.
     * @exception PropertyVetoException when the attempt to set the property is
     * vetoed by a listener
     */
    public final void dilate() throws PropertyVetoException {
        if (membershipFunction != null) {
            membershipFunction.dilate();
        }
    }

    /**
     * Returns the height of this fuzzy set.
     * @return the current height
     */
    public final float getHeight() {
        float height = 0.0f;
        if (membershipFunction != null) {
            height = membershipFunction.getHeight();
        }
        return height;
    }

    /**
     * Checks whether the membership function is convex. A membership function
     * is convex if there is only one change of sign from + to -.
     * @return <code>true</code> if membership function is convex
     * <code>false</code>, otherwise
     */
    public final boolean isConvex() {
        boolean isConvex = false;
        if (membershipFunction != null) {
            isConvex = membershipFunction.isConvex();
        }
        return isConvex;
    }

    /**
     * Checks whether the membership function is normalized. A membership
     * function is normalized if its height is 1.0.
     * @return <code>true</code> if membership function is normalized
     * <code>false</code>, otherwise
     * @see #getHeight
     */
    public final boolean isNormalized() {
        boolean isNormalized = false;
        if (membershipFunction != null) {
            isNormalized = membershipFunction.isNormalized();
        }
        return isNormalized;
    }

    /**
     * Checks whether the fuzzy set fulfills the requirements of a fuzzy interval.
     * The membership function has to be convex and normalized. In addition,
     * the maximum has to lie between a closed interval (<tt>[a,b]</tt> with
     * <tt>a != b</tt>).
     * @return <code>true</code> if the membership function would be a fuzzy
     * interval <code>false</code>, otherwise
     * @see #isNormalized
     * @see #isConvex
     */
    public final boolean isValidFuzzyInterval() {
        if (membershipFunction != null) {
            if (isNormalized() && isConvex()) {
                // Additionally, we are checking whether the maximum is in an interval
                boolean max = false; // maximum not found

                for (Iterator<Float> it = membershipFunction.iterator(); it.hasNext();) {
                    float x = it.next();
                    if (membershipFunction.getDegreeOfMembership(x) == 1.0f) {
                        if (!max) {
                            max = true;
                        } else {
                            return true;
                        }
                    }
                }
            }
        }
        return false;
    }

    /**
     * Checks whether the fuzzy set fulfills the requirements of a fuzzy LR interval.
     * In addition to the requirements for a fuzzy interval the membership
     * function must be represented by two reference function L and R.
     * @return <code>true</code> if the fuzzy set would be a fuzzy LR interval
     * <code>false</code>, otherwise
     * @see #isValidFuzzyInterval
     */
    public final boolean isValidFuzzyLRInterval() {
        if (membershipFunction != null) {
            boolean isValidFuzzyInterval = isValidFuzzyInterval();
            if (isValidFuzzyInterval) {
                // Creating a copy
                MembershipFunction f = (MembershipFunction) membershipFunction.clone();
                f.reduce();
                if (f.size() == 4) {
                    return true;
                }
            }
        }
        return false;
    }

    /**
     * Checks whether the fuzzy set fulfills the requirements of a fuzzy LR number.
     * In addition to the requirements for a fuzzy number the membership
     * function must be represented by two reference function L and R.
     * @return <code>true</code> if the fuzzy set would be a fuzzy LR number
     * <code>false</code>, otherwise
     * @see #isValidFuzzyNumber
     */
    public final boolean isValidFuzzyLRNumber() {
        if (membershipFunction != null) {
            boolean isValidFuzzyNumber = isValidFuzzyNumber();
            if (isValidFuzzyNumber) {
                // Creating a copy
                MembershipFunction f = (MembershipFunction) membershipFunction.clone();
                f.reduce();
                if (f.size() == 3) {
                    return true;
                }
            }
        }
        return false;
    }

    /**
     * Checks whether the fuzzy set fulfills the requirements of a fuzzy number.
     * The membership function has to be convex and normalized. In addition,
     * the maximum is reached at one point.
     * @return <code>true</code> if the fuzzy set would be a fuzzy number
     * <code>false</code>, otherwise
     * @see #isNormalized
     * @see #isConvex
     */
    public final boolean isValidFuzzyNumber() {
        if (membershipFunction != null) {

            if (isNormalized() && isConvex()) {
                // Additionally, we are checking whether the maximum is defined in one place
                boolean max = false; // Maximum not found

                for (Iterator<Float> it = membershipFunction.iterator(); it.hasNext();) {
                    float x = it.next();
                    if (membershipFunction.getDegreeOfMembership(x) == 1.0f) {
                        if (!max) {
                            max = true;
                        } else {
                            return false;
                        }
                    }
                }
                return true;
            }
        }
        return false;
    }

    /**
     * Normalizes this fuzzy set.
     * @exception PropertyVetoException when the attempt to set the property is
     * vetoed by a listener
     */
    public final void normalize() throws PropertyVetoException {
        if (membershipFunction != null) {
            membershipFunction.normalize();
        }
    }

    /**
     * Computes the reciproce fuzzy set.
     * @exception PropertyVetoException when the attempt to set the property is
     * vetoed by a listener
     */
    public final void reciproce() throws PropertyVetoException {
        if (membershipFunction != null) {
            membershipFunction.reciproce();
        }
    }

    /**
     * Adds a <code>PropertyChangeListener</code> to the listener list. The
     * listener is registered for all properties. <p>
     * A <code>PropertyChangeEvent</code> will get fired in response to setting
     * a bound property. <p>
     * @param listener  the <code>PropertyChangeListener</code> to be added
     */
    public final synchronized void addPropertyChangeListener(
            final PropertyChangeListener listener) {
        propertyChangeSupport.addPropertyChangeListener(listener);
    }

    /**
     * Removes a <code>PropertyChangeListener</code> from the listener list.
     * This removes a <code>PropertyChangeListener</code> that was registered
     * for all properties.
     * @param listener  the <code>PropertyChangeListener</code> to be removed
     */
    public final synchronized void removePropertyChangeListener(
            final PropertyChangeListener listener) {
        propertyChangeSupport.removePropertyChangeListener(listener);
    }

    @Override
    public final void propertyChange(final PropertyChangeEvent evt) {
        String prop = evt.getPropertyName();

        if (evt.getSource() == membershipFunction) {

            if (prop.equals(MembershipFunction.DEGREE_OF_MEMBERSHIP_PROPERTY)) {
                try {

                    MembershipFunctionPoint p = (MembershipFunctionPoint) evt.getOldValue();
                    FuzzySet oldFuzzySet = (FuzzySet) this.clone();
                    oldFuzzySet.getMembershipFunction().setDegreeOfMembership(p.getX(), p.getDegreeOfMembership());

                    propertyChangeSupport.firePropertyChange(HEIGHT_PROPERTY,
                            oldFuzzySet.getHeight(), getHeight());

                    propertyChangeSupport.firePropertyChange(CONVEX_PROPERTY,
                            oldFuzzySet.isConvex(), isConvex());

                    propertyChangeSupport.firePropertyChange(NORMALIZED_PROPERTY,
                            oldFuzzySet.isNormalized(), isNormalized());

                    propertyChangeSupport.firePropertyChange(VALID_FUZZY_NUMBER_PROPERTY,
                            oldFuzzySet.isValidFuzzyNumber(), isValidFuzzyNumber());

                    propertyChangeSupport.firePropertyChange(VALID_FUZZY_LR_NUMBER_PROPERTY,
                            oldFuzzySet.isValidFuzzyLRNumber(), isValidFuzzyLRNumber());

                    propertyChangeSupport.firePropertyChange(VALID_FUZZY_INTERVAL_PROPERTY,
                            oldFuzzySet.isValidFuzzyInterval(), isValidFuzzyInterval());

                    propertyChangeSupport.firePropertyChange(VALID_FUZZY_LR_INTERVAL_PROPERTY,
                            oldFuzzySet.isValidFuzzyLRInterval(), isValidFuzzyLRInterval());

                } catch (PropertyVetoException ex) {
                    Logger.getLogger(FuzzyNumber.class.getName()).log(Level.SEVERE, null, ex);
                }
            }
        }
    }

    @Override
    public Object clone() {
        return SerializationUtils.clone(this);
    }

    @Override
    public final boolean equals(Object obj) {
        if (obj == null) {
            return false;
        }
        if (obj == this) {
            return true;
        }
        if (obj.getClass() != getClass()) {
            return false;
        }
        FuzzySet fs = (FuzzySet) obj;
        return new EqualsBuilder().append(this.id, fs.id).append(this.membershipFunction, fs.membershipFunction).isEquals();
    }

    @Override
    public int hashCode() {
        return new HashCodeBuilder(11, 21).append(this.id).append(this.membershipFunction).toHashCode();
    }

    @Override
    public String toString() {
        return toString(true);
    }

    /**
     * Returns a textual representation of the fuzzy set
     *
     * @param withMembershipFunction
     *            <code>true</code> if all points of membership function are also returned,
     *            <code>false</code> otherwise.
     * @return a string representation of the fuzzy set
     * @see MembershipFunction#toString
     */
    public String toString(final boolean withMembershipFunction) {
        if (membershipFunction == null) {
            return FuzzyResourceManager.getString(this,
                    "FUZZY_SET_UNKNOWN_MEMBERSHIP_FUNCTION",
                    new Object[]{
                        id,
                        membershipFunction
                    });
        }
        if (withMembershipFunction) {
            return FuzzyResourceManager.getString(this,
                    "FUZZY_SET_WITH_MEMBERSHIP_FUNCTION",
                    new Object[]{
                        id,
                        membershipFunction.toString()
                    });
        } else {
            Defuzzificator def = new Defuzzificator(Defuzzificator.TYPE_CENTER_OF_AREA);
            return FuzzyResourceManager.getString(this,
                    "FUZZY_SET_WITHOUT_MEMBERSHIP_FUNCTION",
                    new Object[]{
                        id,
                        Float.toString(def.defuzzify(this)),
                        membershipFunction.toString()
                    });
        }
    }
}
